#!/usr/bin/python

from nilmdb.utils.printf import *

import threading
import multiprocessing
import cStringIO
import sys
import os
import signal
import time
import uuid
import psutil
import imp
import traceback

class LogReceiver(object):
    """Spawn a thread that listens to a pipe for log messages,
    and stores them locally."""
    def __init__(self, pipe):
        self.pipe = pipe
        self.log = cStringIO.StringIO()
        self.thread = threading.Thread(target = self.run)
        self.thread.start()

    def run(self):
        while True:
            data = os.read(self.pipe, 65536)
            if not data:
                os.close(self.pipe)
                return
            self.log.write(data)

    def getvalue(self):
        return self.log.getvalue()

    def clear(self):
        self.log = cStringIO.StringIO()

class Process(object):
    """Spawn and manage a process that calls a Python function"""
    def __init__(self, name, function, parameters):
        self.parameters = parameters
        self.start_time = None
        self.name = name

        # Use a pipe for communicating log data
        (rpipe, wpipe) = os.pipe()
        self._log = LogReceiver(rpipe)

        # Start the function in a new process
        self._process = multiprocessing.Process(
            target = self._trampoline, name = name,
            args = (function, rpipe, wpipe, parameters))
        self._process.daemon = True
        self._process.start()

        # Close the writer end of the pipe, get process info
        os.close(wpipe)
        self.start_time = time.time()
        self.pid = str(uuid.uuid1(self._process.pid or 0))

    def _trampoline(self, func, rpipe, wpipe, param): # pragma: no cover
        # No coverage report for this, because it's executed in a subprocess
        """Trampoline function to set up stdio and call the real function."""
        # Close the reader end of the pipe
        os.close(rpipe)

        # Like os.close() but ignores errors
        def tryclose(fd):
            try:
                os.close(fd)
            except OSError:
                pass

        # Remap stdio to go to the pipe.  We do this at the OS level,
        # replacing FDs, so that future spawned processes do the right thing.

        # stdin
        sys.stdin.close()
        tryclose(0)
        fd = os.open(os.devnull, os.O_RDONLY) # 0
        sys.stdin = os.fdopen(fd, 'r', 0)

        # stdout
        sys.stdout.close()
        tryclose(1)
        fd = os.dup(wpipe) # 1
        sys.stdout = os.fdopen(fd, 'w', 0)

        # stdout
        sys.stderr.close()
        tryclose(2)
        fd = os.dup(wpipe) # 2
        sys.stderr = os.fdopen(fd, 'w', 0)

        # Don't need this extra fd
        os.close(wpipe)

        # Ready to go -- call the function, exit when it's done
        func(param)
        sys.exit(0)

    def terminate(self, timeout = 1.0):
        """Terminate a process, and all of its children that are in the same
        process group."""
        # First give it some time to die on its own
        self._process.join(timeout)
        if not self.alive:
            return True

        def getpgid(pid):
            try:
                return os.getpgid(pid)
            except OSError: # pragma: no cover
                return None

        # Find all children
        group = getpgid(self._process.pid)
        main = psutil.Process(self._process.pid)
        allproc = [ main ] + main.get_children(recursive = True)

        # Kill with SIGTERM, if they're still in this process group
        for proc in allproc:
            if getpgid(proc.pid) == group:
                os.kill(proc.pid, signal.SIGTERM)

        # Wait for it to die again
        self._process.join(timeout)
        if not self.alive:
            return True

        # One more try with SIGKILL
        for proc in allproc:
            if getpgid(proc.pid) == group:
                os.kill(proc.pid, signal.SIGKILL)

        # See if it worked
        self._process.join(timeout)
        return not self.alive

    def clear_log(self):
        self._log.clear()

    @property
    def log(self):
        return self._log.getvalue()

    @property
    def alive(self):
        return self._process.is_alive()

    @property
    def exitcode(self):
        return self._process.exitcode

def _exec_user_code(codeargs): # pragma: no cover (runs in subprocess)
    """Execute 'code' as if it were placed into a file and executed"""
    (code, args) = codeargs
    # This is split off into a separate function because the Python3
    # syntax of "exec" triggers a SyntaxError in Python2, if it's within
    # a nested function.
    imp.acquire_lock()
    try:
        module = imp.new_module("__main__")
    finally:
        imp.release_lock()
    module.__file__ = "<user-code>"
    sys.argv = [''] + args
    # Wrap the compile and exec in a try/except so we can format the
    # exception more nicely.
    try:
        codeobj = compile(code, '<user-code>', 'exec',
                          flags = 0, dont_inherit = 1)
        exec(codeobj, module.__dict__, {})
    except Exception:
        try:
            # Pull out the exception
            info = sys.exc_info()
            tblist = traceback.extract_tb(info[2])

            # First entry is probably this code; get rid of it
            if len(tblist) and tblist[0][2] == '_exec_user_code':
                tblist = tblist[1:]

            # Add the user's source code to every line that's missing it
            lines = code.splitlines()
            maxline = len(lines)
            for (n, (name, line, func, text)) in enumerate(tblist):
                if name == '<user-code>' and text is None and line <= maxline:
                    tblist[n] = (name, line, func, lines[line-1].strip())

            # Format it in the usual manner
            out = ['Traceback (most recent call last):\n']
            out.extend(traceback.format_list(tblist))
            out.extend(traceback.format_exception_only(info[0], info[1]))
        finally:
            # Need to explicitly delete traceback object to avoid ref cycle
            del info
        sys.stderr.write("".join(out))
        sys.stderr.flush()
        sys.exit(1)
    sys.exit(0)

class ProcessManager(object):
    """Track and manage a collection of Process objects"""
    def __init__(self):
        self.processes = {}

    def __iter__(self):
        return iter(self.processes.keys())

    def __getitem__(self, key):
        return self.processes[key]

    def run_function(self, procname, function, parameters):
        """Run a Python function that already exists"""
        new = Process(procname, function, parameters)
        self.processes[new.pid] = new
        return new.pid

    def run_code(self, procname, code, args):
        """Evaluate 'code' as if it were placed into a Python file and
        executed.  The arguments will be accessible in the code as
        sys.argv[1:]."""
        return self.run_function(procname, _exec_user_code, (code, args))

    def run_command(self, procname, argv):
        """Execute a command line program"""
        def spwan_user_command(argv): # pragma: no cover (runs in subprocess)
            try:
                maxfd = os.sysconf("SC_OPEN_MAX")
            except Exception:
                maxfd = 256
            os.closerange(3, maxfd)
            try:
                os.chdir("/tmp")
            except OSError:
                pass
            os.execvp(argv[0], argv)
        return self.run_function(procname, spwan_user_command, argv)

    def terminate(self, pid):
        return self.processes[pid].terminate()

    def remove(self, pid):
        del self.processes[pid]
